/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

You may use this sample code for anything you like, it is not covered by the
LGPL like the rest of the engine.
-----------------------------------------------------------------------------
*/
/*
-----------------------------------------------------------------------------
Filename:    ExampleFrameListener.h
Description: Defines an example frame listener which responds to frame events.
This frame listener just moves a specified camera around based on
keyboard and mouse movements.
Mouse:    Freelook
W or Up:  Forward
S or Down:Backward
A:        Step left
D:        Step right
             PgUp:     Move upwards
             PgDown:   Move downwards
             F:        Toggle frame rate stats on/off
			 R:        Render mode
             T:        Cycle texture filtering
                       Bilinear, Trilinear, Anisotropic(8)
             P:        Toggle on/off display of camera position / orientation
-----------------------------------------------------------------------------
*/

#ifndef __KdTreeAppListener_H__
#define __KdTreeAppListener_H__

#include "Ogre.h"
#include "OgreKeyEvent.h"
#include "OgreEventListeners.h"
#include "OgreStringConverter.h"
#include "OgreException.h"

#define VIZ_VIEWPORT_Z_ORDER 10

using namespace Ogre;

struct FrameInfo
{
	FrameInfo(Vector3 pos, Quaternion or, Real time):
	mPosition(pos), mOrientation(or), mElapsedTime(time)
	{ }

	Vector3 mPosition;
	Quaternion mOrientation;
	Real	mElapsedTime;
};

typedef std::list<FrameInfo> FrameList;

class KdTreeAppListener: public FrameListener, public KeyListener
{

public:
	// visualization modes for scene nodes
	enum VizMode
	{
		NODEVIZ_NONE, 
		NODEVIZ_RENDER_NODES, 
		NODEVIZ_RENDER_NODES_AND_CONTENT, 
		NODEVIZ_MODES_NUM
	};

	enum AppState
	{
		AS_NORMAL,
		AS_RECORD,
		AS_PLAYBACK
	};

	struct Options
	{
		Options():
		mDemoMode(false),
		mRotateSpeed(36.0f),
		mMoveSpeed(100.0f),
		mRotationPeriod(30.0f),
		mMaxDepth(12),
		mKT(2.0f),
		mKI(1.0f),
		mRenderMethod("CHC")
		{

		}

		String mSceneManager;
		String mDemoInfileName;
		String mDemoOutfileName;
		String mDemoLogfileName;
		bool mDemoMode;
		Degree mRotateSpeed;
		Real mMoveSpeed;
		Real mRotationPeriod;
		int mMaxDepth;
		Real mKT;
		Real mKI;
		String mRenderMethod;
		String mComment;
	};

    // Constructor takes a RenderWindow because it uses that to determine input context
    KdTreeAppListener(RenderWindow* win, SceneManager* sm, const Options& options,  
		bool useBufferedInputKeys = false, bool useBufferedInputMouse = false);

    virtual ~KdTreeAppListener();

	void initOverlayElement(OverlayElement **elInfo, String ext, 
		String name, int top, String caption);

	void initStatsOverlay();

	void initKdTreeOverlay();

	void showDebugOverlay(bool show);

	void toggleVizCamera();

    virtual bool processUnbufferedKeyInput(const FrameEvent& evt);

    bool processUnbufferedMouseInput(const FrameEvent& evt);

	void moveCamera();

    // Override frameStarted event to process that (don't care about frameEnded)
    bool frameStarted(const FrameEvent& evt);

    bool frameEnded(const FrameEvent& evt);

	void switchMouseMode();

	void switchKeyMode();

	void keyClicked(KeyEvent* e);

	void keyPressed(KeyEvent* e) {};
	void keyReleased(KeyEvent* e) {};


protected:
	// basic
	RenderWindow* mWindow;
	SceneManager *mSceneMgr;
	Options mOptions;
	bool mUseBufferedInputKeys, mUseBufferedInputMouse, mInputTypeSwitchingOn;

	// elements
	SceneNode *mCamNode;
	Camera* mCamera;
	Camera* mTopCam;

	// toggle
	bool mStatsOn;

	bool mVizCamera;
	bool mShowTree;
	bool mFreeMove;
	bool mTopCamFollow;

	// counters
	int mSeqNum;
	unsigned int mNumScreenShots;

	// rendering/texture options
	int mSceneDetailIndex;
	TextureFilterOptions mFiltering;
	int mAniso;

	// chc stats
	Real mDelayedQueriesIssued;
	Real mDelayedTraversedNodes;

	// overlays
	Overlay *mDemoOverlay;
	Overlay *mDebugOverlay;
	Overlay *mKdTreeOverlay;

	OverlayElement *mDemoStatus;

	OverlayElement *mKdTreeMaxDepthInfo;
	OverlayElement *mKTInfo;
	OverlayElement *mKIInfo;
	OverlayElement *mHighlightLevelInfo;
	OverlayElement *mBuildMethodInfo;
	OverlayElement *mRenderMethodInfo;
	OverlayElement *mMovementInfo;
	OverlayElement *mTopCamInfo;

	OverlayElement *mFrustumCulledNodesInfo;
	OverlayElement *mQueryCulledNodesInfo;
	OverlayElement *mTraversedNodesInfo;
	OverlayElement *mHierarchyNodesInfo;
	OverlayElement *mRenderedNodesInfo;
	OverlayElement *mObjectsCountInfo;
	OverlayElement *mQueriesIssuedInfo;

	// input
	EventProcessor* mEventProcessor;
	InputReader* mInputDevice;

	// movement
	Vector3 mTranslateVector;
	Vector3 mVizCamTransVect;
	Radian mRotX, mRotY;
	float mMoveScale;
	Degree mRotScale;
	Radian mDeathAngle;

	// just to stop toggles flipping too fast
	Real mTimeUntilNextToggle ;

	// stuff for walkthrough recording/playback
	//std::ofstream mDemoFramesLog;
	std::list<Real> mDemoFPS;
	Real mTimeUntilNextLogWrite;
	FrameList mFrameList;
	FrameList::iterator mCurrFrame;
	Real mTimeRemaining;
	Real mWaitBeforeDemoStart;

	AppState mAppState;

	void setDemoOverlay();

	void toggleRecord();
	void togglePlayback();

	/** write FPS stats to log file (csv) **/
	void saveLog();

	void saveFrameInfo(Real elapsedTime);

	/** save demo frame data to file **/
	static void saveFrames(const String& filename, FrameList& framelist);
	static void saveFramesAscii(const String& filename, FrameList& framelist);
	static void saveFramesBinary(const String& filename, FrameList& framelist);

	/** load demo frame data from file **/
	static void loadFrames(const String& filename, FrameList& framelist);
	static void loadFramesAscii(const String& filename, FrameList& framelist);
	static void loadFramesBinary(const String& filename, FrameList& framelist);

	void updateStats(void);
};

#endif

